use termion::{clear, cursor, raw::RawTerminal};

const ESTIMATED_FRAME_BIT_SIZE: usize = 11;
use std::io::{Stdout, Write};

use super::theme::{ColorSet, Theme};

const FRAME_CHAR_HORIZONTAL: char = ' ';
const FRAME_CHAR_VERTICAL: char = ' ';

pub enum FrameSize {
    ByAbsolute(u16, u16),
    ByPercent(u16, u16),
}

#[derive(Debug, Clone, Copy)]
pub struct Frame {
    // Both are (w, h)
    start: (u16, u16),
    end: (u16, u16),
    border: u16,
    theme: ColorSet,
}

impl Frame {
    #[inline(always)]
    pub fn sub(
        &self,
        theme: ColorSet,
        size: FrameSize,
        border: u16,
    ) -> Frame {
        let (p_width, p_height) = self.size();
        let sub_size = size.abs_size((
            p_width - self.border * 2,
            p_height - self.border * 2,
        ));
        self.child_centered(sub_size, border, theme)
    }

    #[inline(always)]
    pub fn goto(&self, x: u16, y: u16) -> String {
        let cg = cursor::Goto(
            self.border.max(
                (self.border + self.start.0 + x)
                    .min(self.end.0 - self.border),
            ),
            self.border.max(
                (self.border + self.start.1 + y)
                    .min(self.end.1 - self.border),
            ),
        );
        cg.into()
    }

    #[inline(always)]
    fn goto_internal(&self, x: u16, y: u16) -> String {
        cursor::Goto(
            (self.start.0 + x).min(self.end.0),
            (self.start.1 + y).min(self.end.1),
        )
        .into()
    }

    #[inline(always)]
    fn write_centered_clear(&self, s: &str) -> String {
        let width = self.size().0 as usize;
        let limit = width - self.border as usize * 2;
        let s = if s.len() > limit { &s[..limit] } else { s };
        let len = s.len();
        let base_size = ((width - len) / 2) - self.border as usize;
        format!(
            "{left}{line}{right}",
            left = " ".repeat(base_size + ((width - len) % 2)),
            line = s,
            right = " ".repeat(base_size)
        )
    }

    #[inline(always)]
    fn write_frame_line(&self, y: u16, length: u16) -> String {
        self.goto_internal(0, y)
            + &FRAME_CHAR_HORIZONTAL
                .to_string()
                .repeat(length as usize)
    }

    #[inline(always)]
    fn write_clear_to_end(&self, s: &str) -> String {
        let limit = (self.size().0 - self.border * 2) as usize;
        let s = if s.len() > limit { &s[..limit] } else { s };
        let clear_length = self.size().0 as usize
            - s.len()
            - (self.border * 2) as usize;
        format!("{}{}", s, " ".repeat(clear_length))
    }

    pub fn writeln(&self, s: &str, x: u16, y: u16) -> String {
        let words = s.split('\n').collect::<Vec<&str>>();
        let mut out_words = Vec::with_capacity(words.len());
        for i in 0..words.len() {
            out_words.push(format!(
                "{ret}{str}",
                str = self.write_clear_to_end(words[i]),
                ret = self.goto(x, y + i as u16),
            ));
        }
        out_words.concat()
    }

    pub fn write_centered(&self, s: &str, y: u16) -> String {
        let words = s.split('\n').collect::<Vec<&str>>();
        let mut out_words = Vec::with_capacity(words.len());
        for i in 0..words.len() {
            out_words.push(format!(
                "{ret}{str}",
                ret = self.goto(0, y + i as u16),
                str = self.write_centered_clear(words[i]),
            ));
        }
        out_words.concat()
    }

    #[inline(always)]
    pub fn size(&self) -> (u16, u16) {
        eprintln!(
            "end.0: {}, start.0: {}, end.1: {}, start.1: {}",
            self.end.0, self.start.0, self.end.1, self.start.1
        );
        (self.end.0 - self.start.0, self.end.1 - self.start.1)
    }

    #[inline(always)]
    pub fn from_terminal_size(
        theme: ColorSet,
        width: u16,
    ) -> Result<Self, anyhow::Error> {
        let term = termion::terminal_size()?;
        Ok(Self::from_bottom_right(term, term, width, theme))
    }

    #[inline(always)]
    fn child_centered(
        &self,
        (pos_w, pos_h): (u16, u16),
        border: u16,
        theme: ColorSet,
    ) -> Frame {
        let parent_offset = self.border / 2;
        let (parent_w, parent_h) = self.size();
        let (parent_w, parent_h) =
            (parent_w - self.border, parent_h - self.border);
        let start = (
            1.max(
                (parent_w - pos_w) / 2
                    + ((parent_w - pos_w) % 2)
                    + parent_offset,
            ),
            1.max(
                (parent_h - pos_h) / 2
                    + ((parent_h - pos_h) % 2)
                    + parent_offset,
            ),
        );
        let frame = Self {
            end: (start.0 + pos_w, start.1 + pos_h),
            start,
            border,
            theme,
        };
        frame
    }

    #[inline(always)]
    fn from_bottom_right(
        (pos_w, pos_h): (u16, u16),
        (term_w, term_h): (u16, u16),
        border: u16,
        theme: ColorSet,
    ) -> Self {
        let start = (1.max(term_w - pos_w), 1.max(term_h - pos_h));
        let frame = Self {
            end: (start.0 + pos_w, start.1 + pos_h),
            start,
            border,
            theme,
        };
        if frame.start.0 > frame.end.0 || frame.start.1 > frame.end.1
        {
            eprintln!(
                "pos_w {}, pos_h {}, term_w {}, term_h {}, border {}",
                pos_w, pos_h, term_w, term_h, border
            );
            panic!(
                "start.0: {} end.0: {}, start.1: {}, end.1: {}",
                frame.start.0,
                frame.end.0,
                frame.start.1,
                frame.end.1
            );
        }

        frame
    }

    pub fn frame_str(&self, body_theme: String) -> String {
        if self.border == 0 {
            return body_theme + &clear::All.to_string();
        }
        self.goto_internal(0, 0);
        let (w_len, h_len) = self.size();
        let width_str = FRAME_CHAR_HORIZONTAL
            .to_string()
            .repeat(w_len as usize + 1);
        let mut frame = Vec::with_capacity(h_len as usize + 4);
        let frame_theme = self.theme.to_string();
        let body_clear =
            body_theme.clone() + &" ".repeat(w_len as usize);
        frame.push(self.theme.to_string());

        for y in 0..self.border {
            frame.push(self.write_frame_line(y, w_len));
        }
        for y in self.border..h_len - self.border {
            frame.push(format!(
                "{left}{body_clear}{frame_theme}{left}{char}{right}{char}",
                body_clear = &body_clear,
                frame_theme = &frame_theme,
                left = self.goto_internal(0, y),
                right = self.goto_internal(w_len - self.border, y),
                char = FRAME_CHAR_VERTICAL
                    .to_string()
                    .repeat(self.border as usize),
            ));
        }
        for y in h_len - self.border..h_len {
            frame.push(self.write_frame_line(y, w_len));
        }
        frame.push(self.goto(0, 0));
        frame.push(body_theme);
        frame.concat()
    }
}

impl FrameSize {
    #[inline(always)]
    fn abs_size(
        &self,
        (term_width, term_height): (u16, u16),
    ) -> (u16, u16) {
        match self {
            FrameSize::ByAbsolute(w, h) => {
                ((*w).min(term_width), (*h).min(term_height))
            }
            FrameSize::ByPercent(w, h) => {
                // term_height            = 100%
                // x                      =   h%
                //
                // term_height * h / 100  =    x
                (
                    term_width * (*w).min(100) / 100,
                    term_height * (*h).min(100) / 100,
                )
            }
        }
    }
}
